import logging, re, database, datetime, os
from dateutil import tz


log = logging.getLogger(__name__)
STRING = re.compile(r'^[\w ]+$')
STRING_NO_SPACE = re.compile(r'^[\w]+$')
TZ = os.getenv('TZ', 'UTC')


def convert_utc_to_local(utc_dt:str, local_tz:str) -> str:
    utc_dt = datetime.datetime.strptime(utc_dt, '%Y-%m-%d %H:%M:%S')

    log.debug(f"Converting UTC datetime {utc_dt} to local timezone {local_tz}")
    if not isinstance(utc_dt, datetime.datetime):
        raise ValueError("utc_dt must be a datetime object")
    
    if not isinstance(local_tz, str):
        raise ValueError("local_tz must be a string")
    
    local_tz = tz.gettz(local_tz)
    if local_tz is None:
        raise ValueError(f"Invalid timezone: {local_tz}")
    
    local_dt = utc_dt.replace(tzinfo=tz.tzutc()).astimezone(local_tz)
    log.debug(f"Converted datetime: {local_dt}")
    return local_dt.strftime('%Y-%m-%d %H:%M:%S')


def validate_string(input_string:str, allowed_regex:re.Pattern=STRING, max_length:int=None, min_length:int=None) -> bool:
    if max_length and len(input_string) > max_length:
        return False
    if min_length and len(input_string) < min_length:
        return False
    if not re.match(allowed_regex, input_string):
        return False
    return True


def validate_board_name(board_name:str) -> bool:
    log.debug(f"Validating board name: {board_name}")
    return validate_string(board_name, allowed_regex=STRING_NO_SPACE, max_length=20, min_length=3)


def validate_username(username:str) -> bool:
    log.debug(f"Validating username: {username}")
    return validate_string(username, allowed_regex=STRING_NO_SPACE, max_length=20, min_length=3)


class data_converter:
    def __init__(self, db:database.Database):
        self.db = db

    def board_to_dict(self, board):
        return {
            'id': board[0],
            'name': board[1],
            'description': board[2],
            'created_at': convert_utc_to_local(board[3], 'Europe/London'),
            'owner_id': board[4]
        }
    
    def post_to_dict(self, post, show_reference:bool=True):
        log.debug(f"Converting post with id ({post[0]}) to dict")
        data = {
            'id': post[0],
            'content': post[3],
            'short_content': post[3][:50] + '...' if len(post[3]) > 50 else post[3],
            'created_at': convert_utc_to_local(post[6], 'Europe/London'),
            'type': post[5],
        }

        # Get the user who created the post
        user = self.db.execute_query("SELECT * FROM users WHERE id = ?", (post[1],), fetch_type=database.FETCHONE)
        if user:
            data['user'] = {
                'id': user[0],
                'name': user[1],
                'about': user[3],
                'avatar_url': '/users/' + str(user[0]) + '/avatar' if user[5] else None,
                'perms': user[4],
            }
        else:
            data['user'] = None

        # Get the board where the post was created
        board = self.db.execute_query("SELECT * FROM boards WHERE id = ?", (post[2],), fetch_type=database.FETCHONE)
        if board:
            data['board'] = {
                'id': board[0],
                'name': board[1],
                'description': board[2],
                'created_at': convert_utc_to_local(board[3], 'Europe/London'),
                'owner_id': board[4]
            }
        else:
            data['board'] = None
    
        # Get the reference post if it exists
        if post[4]:
            reference_post = self.db.execute_query("SELECT * FROM posts WHERE id = ?", (post[4],), fetch_type=database.FETCHONE)
            if reference_post:
                data['reference'] = {
                    'id': reference_post[0],
                    'content': reference_post[3],
                    'short_content': reference_post[3][:25] + '...' if len(reference_post[3]) > 25 else reference_post[3],
                    'created_at': convert_utc_to_local(reference_post[6], 'Europe/London'),
                    'type': reference_post[5],
                    'show': show_reference
                }
            else:
                data['reference'] = None
        else:
            data['reference'] = None

        # Get the attachments for the post
        attachments = self.db.execute_query("SELECT * FROM attachments WHERE post_id = ?", (post[0],), fetch_type=database.FETCHALL)
        data['attachments'] = []
        for attachment in attachments:
            data['attachments'].append({
                'id': attachment[0],
                'file_name': attachment[3],
                'created_at': convert_utc_to_local(attachment[5], 'Europe/London'),
                'url': '/attachment/' + str(attachment[0]),
                'type': attachment[4].split('/')[0],
                'mime_type': attachment[4]
            })

        # Get the number of replies to the post
        replies = self.db.execute_query("SELECT COUNT(*) FROM posts WHERE reference = ?", (post[0],), fetch_type=database.FETCHONE)
        data['replies'] = replies[0] if replies else 0
        log.debug(f"Post converted to dict")
        return data
    

    def user_to_dict(self, user):
        return {
            'id': user[0],
            'name': user[1],
            'about': user[3],
            'avatar_url': '/users/' + str(user[0]) + '/avatar' if user[5] else None,
            'perms': user[4],
            'created_at': convert_utc_to_local(user[6], 'Europe/London'),
            'avatar_type': user[8],
        }